@@ -4,34 +4,52 @@ import (
 	"os"
 	"strings"
 
-	"github.com/git-lfs/git-lfs/git"
-	"github.com/git-lfs/git-lfs/lfs"
-	"github.com/git-lfs/git-lfs/tools/humanize"
+	"github.com/git-lfs/git-lfs/v3/git"
+	"github.com/git-lfs/git-lfs/v3/lfs"
+	"github.com/git-lfs/git-lfs/v3/tools/humanize"
 	"github.com/spf13/cobra"
 )
 
 var (
-	longOIDs           = false
-	lsFilesScanAll     = false
-	lsFilesScanDeleted = false
-	lsFilesShowSize    = false
-	debug              = false
+	longOIDs            = false
+	lsFilesScanAll      = false
+	lsFilesScanDeleted  = false
+	lsFilesShowSize     = false
+	lsFilesShowNameOnly = false
+	debug               = false
 )
 
 func lsFilesCommand(cmd *cobra.Command, args []string) {
-	requireInRepo()
+	setupRepository()
 
 	var ref string
-
-	if len(args) == 1 {
+	var otherRef string
+	var scanRange = false
+	if len(args) > 0 {
 		if lsFilesScanAll {
 			Exit("fatal: cannot use --all with explicit reference")
+		} else if args[0] == "--all" {
+			// Since --all is a valid argument to "git rev-parse",
+			// if we try to give it to git.ResolveRef below, we'll
+			// get an unexpected result.
+			//
+			// So, let's check early that the caller invoked the
+			// command correctly.
+			Exit("fatal: did you mean \"git lfs ls-files --all --\" ?")
 		}
+
 		ref = args[0]
+		if len(args) > 1 {
+			if lsFilesScanDeleted {
+				Exit("fatal: cannot use --deleted with reference range")
+			}
+			otherRef = args[1]
+			scanRange = true
+		}
 	} else {
 		fullref, err := git.CurrentRef()
 		if err != nil {
-			ref = git.RefBeforeFirstCommit
+			ref = git.EmptyTree()
 		} else {
 			ref = fullref.Sha
 		}
@@ -44,13 +62,13 @@ func lsFilesCommand(cmd *cobra.Command, args []string) {
 
 	seen := make(map[string]struct{})
 
-	gitscanner := lfs.NewGitScanner(func(p *lfs.WrappedPointer, err error) {
+	gitscanner := lfs.NewGitScanner(cfg, func(p *lfs.WrappedPointer, err error) {
 		if err != nil {
 			Exit("Could not scan for Git LFS tree: %s", err)
 			return
 		}
 
-		if !lsFilesScanAll {
+		if !lsFilesScanAll && !scanRange {
 			if _, ok := seen[p.Name]; ok {
 				return
 			}
@@ -73,6 +91,9 @@ func lsFilesCommand(cmd *cobra.Command, args []string) {
 				p.Version)
 		} else {
 			msg := []string{p.Oid[:showOidLen], lsFilesMarker(p), p.Name}
+			if lsFilesShowNameOnly {
+				msg = []string{p.Name}
+			}
 			if lsFilesShowSize {
 				size := humanize.FormatBytes(uint64(p.Size))
 				msg = append(msg, "("+size+")")
@@ -86,10 +107,17 @@ func lsFilesCommand(cmd *cobra.Command, args []string) {
 	defer gitscanner.Close()
 
 	includeArg, excludeArg := getIncludeExcludeArgs(cmd)
-	gitscanner.Filter = buildFilepathFilter(cfg, includeArg, excludeArg)
+	gitscanner.Filter = buildFilepathFilter(cfg, includeArg, excludeArg, false)
 
-	if err := gitscanner.ScanIndex(ref, nil); err != nil {
-		Exit("Could not scan for Git LFS index: %s", err)
+	if len(args) == 0 {
+		// Only scan the index when "git lfs ls-files" was invoked with
+		// no arguments.
+		//
+		// Do so to avoid showing "mixed" results, e.g., ls-files output
+		// from a specific historical revision, and the index.
+		if err := gitscanner.ScanIndex(ref, nil); err != nil {
+			Exit("Could not scan for Git LFS index: %s", err)
+		}
 	}
 	if lsFilesScanAll {
 		if err := gitscanner.ScanAll(nil); err != nil {
@@ -99,6 +127,8 @@ func lsFilesCommand(cmd *cobra.Command, args []string) {
 		var err error
 		if lsFilesScanDeleted {
 			err = gitscanner.ScanRefWithDeleted(ref, nil)
+		} else if scanRange {
+			err = gitscanner.ScanRefRange(otherRef, ref, nil)
 		} else {
 			err = gitscanner.ScanTree(ref)
 		}
@@ -127,6 +157,7 @@ func init() {
 	RegisterCommand("ls-files", lsFilesCommand, func(cmd *cobra.Command) {
 		cmd.Flags().BoolVarP(&longOIDs, "long", "l", false, "")
 		cmd.Flags().BoolVarP(&lsFilesShowSize, "size", "s", false, "")
+		cmd.Flags().BoolVarP(&lsFilesShowNameOnly, "name-only", "n", false, "")
 		cmd.Flags().BoolVarP(&debug, "debug", "d", false, "")
 		cmd.Flags().BoolVarP(&lsFilesScanAll, "all", "a", false, "")
 		cmd.Flags().BoolVar(&lsFilesScanDeleted, "deleted", false, "")
