@@ -10,12 +10,12 @@ import (
 	"strings"
 	"sync"
 
-	"github.com/git-lfs/git-lfs/errors"
-	"github.com/git-lfs/git-lfs/git"
-	"github.com/git-lfs/git-lfs/lfs"
-	"github.com/git-lfs/git-lfs/tasklog"
-	"github.com/git-lfs/git-lfs/tools"
-	"github.com/git-lfs/git-lfs/tq"
+	"github.com/git-lfs/git-lfs/v3/errors"
+	"github.com/git-lfs/git-lfs/v3/git"
+	"github.com/git-lfs/git-lfs/v3/lfs"
+	"github.com/git-lfs/git-lfs/v3/tasklog"
+	"github.com/git-lfs/git-lfs/v3/tools"
+	"github.com/git-lfs/git-lfs/v3/tq"
 	"github.com/rubyist/tracerx"
 )
 
@@ -31,12 +31,19 @@ func uploadForRefUpdates(ctx *uploadContext, updates []*git.RefUpdate, pushAll b
 	}()
 
 	verifyLocksForUpdates(ctx.lockVerifier, updates)
+	rightSides := make([]string, 0, len(updates))
+	for _, update := range updates {
+		right := update.Right().Sha
+		if update.LeftCommitish() != right {
+			rightSides = append(rightSides, right)
+		}
+	}
 	for _, update := range updates {
 		// initialized here to prevent looped defer
 		q := ctx.NewQueue(
 			tq.RemoteRef(update.Right()),
 		)
-		err := uploadLeftOrAll(gitscanner, ctx, q, update, pushAll)
+		err := uploadLeftOrAll(gitscanner, ctx, q, rightSides, update, pushAll)
 		ctx.CollectErrors(q)
 
 		if err != nil {
@@ -47,14 +54,19 @@ func uploadForRefUpdates(ctx *uploadContext, updates []*git.RefUpdate, pushAll b
 	return nil
 }
 
-func uploadLeftOrAll(g *lfs.GitScanner, ctx *uploadContext, q *tq.TransferQueue, update *git.RefUpdate, pushAll bool) error {
+func uploadLeftOrAll(g *lfs.GitScanner, ctx *uploadContext, q *tq.TransferQueue, bases []string, update *git.RefUpdate, pushAll bool) error {
 	cb := ctx.gitScannerCallback(q)
 	if pushAll {
 		if err := g.ScanRefWithDeleted(update.LeftCommitish(), cb); err != nil {
 			return err
 		}
 	} else {
-		if err := g.ScanLeftToRemote(update.LeftCommitish(), cb); err != nil {
+		left := update.LeftCommitish()
+		right := update.Right().Sha
+		if left == right {
+			right = ""
+		}
+		if err := g.ScanMultiRangeToRemote(left, bases, cb); err != nil {
 			return err
 		}
 	}
@@ -100,7 +112,7 @@ func newUploadContext(dryRun bool) *uploadContext {
 		uploadedOids: tools.NewStringSet(),
 		gitfilter:    lfs.NewGitFilter(cfg),
 		lockVerifier: newLockVerifier(manifest),
-		allowMissing: cfg.Git.Bool("lfs.allowincompletepush", true),
+		allowMissing: cfg.Git.Bool("lfs.allowincompletepush", false),
 		missing:      make(map[string]string),
 		corrupt:      make(map[string]string),
 		otherErrs:    make([]error, 0),
@@ -111,7 +123,9 @@ func newUploadContext(dryRun bool) *uploadContext {
 		sink = ioutil.Discard
 	}
 
-	ctx.logger = tasklog.NewLogger(sink)
+	ctx.logger = tasklog.NewLogger(sink,
+		tasklog.ForceProgress(cfg.ForceProgress()),
+	)
 	ctx.meter = buildProgressMeter(ctx.DryRun, tq.Upload)
 	ctx.logger.Enqueue(ctx.meter)
 	ctx.committerName, ctx.committerEmail = cfg.CurrentCommitter()
@@ -144,7 +158,7 @@ func (c *uploadContext) addScannerError(err error) {
 }
 
 func (c *uploadContext) buildGitScanner() (*lfs.GitScanner, error) {
-	gitscanner := lfs.NewGitScanner(nil)
+	gitscanner := lfs.NewGitScanner(cfg, nil)
 	gitscanner.FoundLockable = func(n string) { c.lockVerifier.LockedByThem(n) }
 	gitscanner.PotentialLockables = c.lockVerifier
 	return gitscanner, gitscanner.RemoteForPush(c.Remote)
@@ -185,7 +199,7 @@ func (c *uploadContext) prepareUpload(unfiltered ...*lfs.WrappedPointer) []*lfs.
 	for _, p := range unfiltered {
 		// object already uploaded in this process, or we've already
 		// seen this OID (see above), skip!
-		if uniqOids.Contains(p.Oid) || c.HasUploaded(p.Oid) {
+		if uniqOids.Contains(p.Oid) || c.HasUploaded(p.Oid) || p.Size == 0 {
 			continue
 		}
 		uniqOids.Add(p.Oid)
@@ -244,7 +258,7 @@ func (c *uploadContext) UploadPointers(q *tq.TransferQueue, unfiltered ...*lfs.W
 			ExitWithError(err)
 		}
 
-		q.Add(t.Name, t.Path, t.Oid, t.Size)
+		q.Add(t.Name, t.Path, t.Oid, t.Size, t.Missing, nil)
 		c.SetUploaded(p.Oid)
 	}
 }
@@ -289,6 +303,11 @@ func (c *uploadContext) ReportErrors() {
 		}
 
 		if !c.allowMissing {
+			pushMissingHint := []string{
+				"hint: Your push was rejected due to missing or corrupt local objects.",
+				"hint: You can disable this check with: 'git config lfs.allowincompletepush true'",
+			}
+			Print(strings.Join(pushMissingHint, "\n"))
 			os.Exit(2)
 		}
 	}
@@ -331,6 +350,8 @@ var (
 )
 
 func (c *uploadContext) uploadTransfer(p *lfs.WrappedPointer) (*tq.Transfer, error) {
+	var missing bool
+
 	filename := p.Name
 	oid := p.Oid
 
@@ -340,40 +361,38 @@ func (c *uploadContext) uploadTransfer(p *lfs.WrappedPointer) (*tq.Transfer, err
 	}
 
 	if len(filename) > 0 {
-		if err = c.ensureFile(filename, localMediaPath); err != nil && !errors.IsCleanPointerError(err) {
+		if missing, err = c.ensureFile(filename, localMediaPath, oid); err != nil && !errors.IsCleanPointerError(err) {
 			return nil, err
 		}
 	}
 
 	return &tq.Transfer{
-		Name: filename,
-		Path: localMediaPath,
-		Oid:  oid,
-		Size: p.Size,
+		Name:    filename,
+		Path:    localMediaPath,
+		Oid:     oid,
+		Size:    p.Size,
+		Missing: missing,
 	}, nil
 }
 
 // ensureFile makes sure that the cleanPath exists before pushing it.  If it
 // does not exist, it attempts to clean it by reading the file at smudgePath.
-func (c *uploadContext) ensureFile(smudgePath, cleanPath string) error {
+func (c *uploadContext) ensureFile(smudgePath, cleanPath, oid string) (bool, error) {
 	if _, err := os.Stat(cleanPath); err == nil {
-		return nil
+		return false, nil
 	}
 
 	localPath := filepath.Join(cfg.LocalWorkingDir(), smudgePath)
 	file, err := os.Open(localPath)
 	if err != nil {
-		if c.allowMissing {
-			return nil
-		}
-		return err
+		return !c.allowMissing, nil
 	}
 
 	defer file.Close()
 
 	stat, err := file.Stat()
 	if err != nil {
-		return err
+		return false, err
 	}
 
 	cleaned, err := c.gitfilter.Clean(file, file.Name(), stat.Size(), nil)
@@ -382,9 +401,9 @@ func (c *uploadContext) ensureFile(smudgePath, cleanPath string) error {
 	}
 
 	if err != nil {
-		return err
+		return false, err
 	}
-	return nil
+	return false, nil
 }
 
 // supportsLockingAPI returns whether or not a given url is known to support
