@@ -8,11 +8,11 @@ import (
 	"strings"
 	"sync"
 
-	"github.com/git-lfs/git-lfs/errors"
-	"github.com/git-lfs/git-lfs/filepathfilter"
-	"github.com/git-lfs/git-lfs/git"
-	"github.com/git-lfs/git-lfs/tasklog"
-	"github.com/git-lfs/gitobj"
+	"github.com/git-lfs/git-lfs/v3/errors"
+	"github.com/git-lfs/git-lfs/v3/filepathfilter"
+	"github.com/git-lfs/git-lfs/v3/git"
+	"github.com/git-lfs/git-lfs/v3/tasklog"
+	"github.com/git-lfs/gitobj/v2"
 )
 
 // Rewriter allows rewriting topologically equivalent Git histories
@@ -158,10 +158,12 @@ var (
 		}
 	}
 
-	// WithLoggerto logs updates caused by the *git/githistory.Rewriter to
+	// WithLoggerTo logs updates caused by the *git/githistory.Rewriter to
 	// the given io.Writer "sink".
-	WithLoggerTo = func(sink io.Writer) rewriterOption {
-		return WithLogger(tasklog.NewLogger(sink))
+	WithLoggerTo = func(sink io.Writer, forceProgress bool) rewriterOption {
+		return WithLogger(tasklog.NewLogger(sink,
+			tasklog.ForceProgress(forceProgress),
+		))
 	}
 
 	// WithLogger logs updates caused by the *git/githistory.Rewriter to the
@@ -224,7 +226,7 @@ func (r *Rewriter) Rewrite(opt *RewriteOptions) ([]byte, error) {
 	if len(opt.ObjectMapFilePath) > 0 {
 		objectMapFile, err = os.OpenFile(opt.ObjectMapFilePath, os.O_RDWR|os.O_CREATE|os.O_EXCL, 0666)
 		if err != nil {
-			return nil, fmt.Errorf("Could not create object map file: %v", err)
+			return nil, fmt.Errorf("could not create object map file: %v", err)
 		}
 		defer objectMapFile.Close()
 	}
@@ -382,7 +384,8 @@ func (r *Rewriter) rewriteTree(commitOID []byte, treeOID []byte, path string,
 		}
 
 		if cached := r.uncacheEntry(entry); cached != nil {
-			entries = append(entries, copyEntry(cached))
+			entries = append(entries, copyEntryMode(cached,
+				entry.Filemode))
 			continue
 		}
 
@@ -434,6 +437,13 @@ func copyEntry(e *gitobj.TreeEntry) *gitobj.TreeEntry {
 	}
 }
 
+func copyEntryMode(e *gitobj.TreeEntry, mode int32) *gitobj.TreeEntry {
+	copied := copyEntry(e)
+	copied.Filemode = mode
+
+	return copied
+}
+
 func (r *Rewriter) allows(typ gitobj.ObjectType, abs string) bool {
 	switch typ {
 	case gitobj.BlobObjectType:
@@ -539,7 +549,7 @@ func (r *Rewriter) refsToMigrate() ([]*git.Ref, error) {
 
 	var local []*git.Ref
 	for _, ref := range refs {
-		if ref.Type == git.RefTypeRemoteBranch || ref.Type == git.RefTypeRemoteTag {
+		if ref.Type == git.RefTypeRemoteBranch {
 			continue
 		}
 
